# Some tests of the behavior of the almanac `_intersection()` function.

def _q(a, b, c, sign):
    # when a x^2 + b x + c = 0
    from numpy import sqrt
    # print('doing quadratic with:', a, b, c)
    print('---')
    print('root (+):', (-b + sqrt(b*b - 4*a*c)) / 2*a)
    print('root (-):', (-b - sqrt(b*b - 4*a*c)) / 2*a)
    print('root alt (+):', - 2*c / (b + sqrt(b*b - 4*a*c)))
    print('root alt (-):', - 2*c / (b - sqrt(b*b - 4*a*c)))
    print('sign:', sign)
    return - 2*c / (b + sign * sqrt(b*b - 4*a*c))

def _intersection(a0, a1, v0, v1):
    # Return the time at which a curve reaches a=0, given its position
    # and velocity a0, v0 at time 0.0 and a1, v1 at time 1.0.
    #
    # (overdetermined, so, ignores v1)
    # print('intersection with:', a0, a1, v0, v1)
    # print('k would be:', 2 * (a1 - a0 - v0))
    sign = 1 - 2 * (a0 > a1)
    tx = _q(a1 - a0 - v0, v0, a0, sign)
    # print('tx =', tx)
    return tx

def test_intersection():
    for v in 0.9, 1, 1.1:
        print(v)
        assert _intersection(0, 4, v, v) == 0.0
        assert _intersection(-5, 0, v, v) == 1.0

    assert _intersection(-1, 3, 4, 4) == 0.25
    assert _intersection(-3, 1, 4, 4) == 0.75

    # Increasing, with other intersection far to the right.

    t = _intersection(-5, 5, 10.1, 9.9)
    assert t > 0.49
    assert t < 0.50

    # Increasing, with other intersection far to the left.

    t = _intersection(-5, 5, 9.9, 10.1)
    assert t > 0.50
    assert t < 0.51

    # Decreasing, with other intersection far to the right.

    t = _intersection(5, -5, -10.1, -9.9)
    assert t > 0.49
    assert t < 0.50

    # Decreasing, with other intersection far to the left.

    t = _intersection(5, -5, -9.9, -10.1)
    assert t > 0.50
    assert t < 0.51



#!/usr/bin/env python3
"""Plot a diagram of a difficult search for Moonrise.

The `design/test_sunrise_moonrise.py` finishes with a search for high
latitude (70°N) moonrises, and prints out the worst miss with a negative
elevation versus the horizon, and the worst positive:

Compute time: 0.163 seconds
Altitude vs horizon: min -5.176029 arcseconds,  max 2.819515 arcseconds
Worst undershot is at index 112:
  2023-08-16 00:47:55.324736  -00deg 48' 46.517584"  -5.176933 arcseconds
  2023-08-16 00:47:55.325735  -00deg 48' 46.516681"  -5.176029 arcseconds
  2023-08-16 00:47:55.326735  -00deg 48' 46.515778"  -5.175126 arcseconds
Worst overshot is at index 70:
  2023-05-26 06:09:37.877960  -00deg 48' 42.952997"   2.818692 arcseconds
  2023-05-26 06:09:37.878960  -00deg 48' 42.952173"   2.819516 arcseconds
  2023-05-26 06:09:37.879960  -00deg 48' 42.951349"   2.820339 arcseconds

Each line is one millisecond apart.  In each case we really wanted the
search to find the millisecond where the arcseconds flip from negative
to positive, but missed.

Why?

This script takes the 2023-08-16 case shown above, instruments the .at()
function to record the calls made by the search for moonrise, and plots
the result so we can see what's happening.  I'm planning to keep this
script around in the repository, so that if any future users find fault
with the new risings routine, I can quickly adapt this script to their
case to see what's going on.

"""
from datetime import date, datetime, timedelta
from math import atan, degrees, tau
from skyfield import VERSION, almanac, api, functions
from skyfield.api import load, Topos, wgs84, N, S, E, W
from skyfield.nutationlib import iau2000b

# Avoid 'This probably means that Tcl wasn't installed properly' error.
import matplotlib
matplotlib.use('Agg')

import numpy as np
import matplotlib.pyplot as plt

# Tweak these.

TARGET = 'moon'
YEAR_MONTH_DAY = 2023, 8, 16
ZOOM_ARCSECONDS = None

# You will get a diagram in return.

ts = load.timescale()
t0 = ts.utc(*YEAR_MONTH_DAY) #- 0.1
t1 = t0 + 1.0 #+ 1.1

eph = load('de421.bsp')
earth = eph['earth']
target = eph[TARGET]
observer = earth + wgs84.latlon(70, 0)

trace = []
def tracing_at(t):
    trace.append(t)
    return real_at(t)

real_at = observer.at
observer.at = tracing_at

final_t, y = almanac.find_risings(observer, target, t0, t1)

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
fig, ax = plt.subplots()

# Drop trace[0], since it's different: it's the initial big-picture
# search for intervals that have risings in them at all, which are then
# filtered out and refined through the subsequent searches.
times = trace[1:]

# And add the final answer back in as though it were a final iteration.
times.append(final_t)

for i, t in enumerate(times, 1):
    alt, az, distance = observer.at(t).observe(target).apparent().altaz()
    print(t.utc_strftime('%Y-%m-%d %H:%M:%S.%f'), alt, az)
    ax.plot(az.degrees, alt.degrees, 'o')
    ax.annotate(f'{i}', (az.degrees, alt.degrees), (-1 + 2 * (i%2), 1),
                textcoords='offset fontsize')

h = almanac.build_horizon_function(target)
horizon_degrees = h(distance) / tau * 360.0
horizon_degrees = horizon_degrees[0]
print(f'Horizon: {horizon_degrees}°')
ax.axhline(horizon_degrees)

miss_degrees = alt.degrees[0] - horizon_degrees
miss_arcseconds = miss_degrees * 3600.0
print(f'Error: {miss_arcseconds} arcseconds')

if ZOOM_ARCSECONDS is not None:
    ax.set_ylim(
        alt.degrees[0] - ZOOM_ARCSECONDS / 3600.0,
        alt.degrees[0] + ZOOM_ARCSECONDS / 3600.0,
    )

one_hour = 1.0 / 24.0
one_minute = one_hour / 60
t0 = ts.tt_jd(min(t.tt[0] for t in times) - 10*one_minute)
t1 = ts.tt_jd(max(t.tt[0] for t in times) + 10*one_minute)
t = ts.linspace(t0, t1, 500)
alt, az, distance = observer.at(t).observe(target).apparent().altaz()
ax.plot(az.degrees, alt.degrees, '-')

ax.grid()
fig.savefig('tmp.png')
print('Wrote tmp.png')



import os
from skyfield.api import load, load_file

def _data_path(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)

# Test file generated with:
# python -m jplephem excerpt 1969/07/29 1969/07/30 de441.bsp de441-1969.bsp

def test_multiple_non_overlapping_segments_per_target():
    ts = load.timescale()
    t = ts.utc(1969, 7, [28, 29, 30, 31])
    eph = load_file(_data_path('./de441-1969.bsp'))

    pluto = eph['pluto barycenter']
    assert type(pluto).__name__ == 'Stack'
    assert len(pluto.segments) == 2

    assert str(pluto.at(t).xyz) == (
        '[[-30.47352052 -30.47318972 -30.47285863 -30.47252727]\n'
        ' [ -0.96671442  -0.96985773  -0.97300104  -0.97614434]\n'
        ' [  8.87919555   8.87811509   8.87703455   8.87595392]] au'
    )

    # Does the Stack correctly offer its .ephemeris to apparent()?

    pluto.at(t).observe(pluto).apparent()

    # TODO: SSB.at(t).observe() fails the above test.
